@@ -27,10 +27,11 @@ Manage GitHub repositories with ``repo-helper``.
 #
 
 # stdlib
+import tempfile
 from base64 import b64encode
 from contextlib import contextmanager
 from getpass import getpass
-from typing import Optional, Union
+from typing import Iterator, Optional, Union
 
 # 3rd party
 import click
@@ -38,17 +39,21 @@ from apeye import RequestsURL
 from consolekit.input import confirm
 from consolekit.terminal_colours import Fore, resolve_color_default
 from consolekit.utils import abort
+from domdf_python_tools.paths import PathPlus
 from domdf_python_tools.stringlist import DelimitedList
 from domdf_python_tools.typing import PathLike
 from dulwich.errors import NotGitRepository
 from dulwich.porcelain import fetch
 from github import Github, GithubException
 from github.AuthenticatedUser import AuthenticatedUser
+from github.ContentFile import ContentFile
 from github.GithubException import UnknownObjectException
 from github.Organization import Organization
 from github.Repository import Repository
 from nacl import encoding, public  # type: ignore
 from repo_helper.core import RepoHelper
+from repo_helper.files.ci_cd import ActionsManager, platform_ci_names
+from repo_helper.utils import set_gh_actions_versions
 from southwark.repo import Repo
 from typing_extensions import NoReturn
 
@@ -60,14 +65,16 @@ from repo_helper_github.options import token_option, version_callback
 __author__: str = "Dominic Davis-Foster"
 __copyright__: str = "2020 Dominic Davis-Foster"
 __license__: str = "MIT License"
-__version__: str = "0.3.0"
+__version__: str = "0.4.0"
 __email__: str = "dominic@davis-foster.co.uk"
 
 __all__ = [
 		"github_command",
 		"echo_rate_limit",
 		"GitHubManager",
+		"IsolatedGitHubManager",
 		"encrypt_secret",
+		"compile_required_checks",
 		]
 
 
@@ -112,7 +119,7 @@ class GitHubManager(RepoHelper):
 	:param verbose: Whether to show information on the GitHub API rate limit.
 	:param colour: Whether to use coloured output.
 
-	.. versionchanged:: 0.3.0
+	.. versionchanged:: 0.4.0
 
 		Added the ``verbose`` and ``colour`` options.
 	"""  # noqa: D400
@@ -124,14 +131,14 @@ class GitHubManager(RepoHelper):
 	"""
 	Whether to show information on the GitHub API rate limit.
 
-	.. versionadded: 0.3.0
+	.. versionadded: 0.4.0
 	"""
 
 	colour: Optional[bool]
 	"""
 	Whether to use coloured output.
 
-	.. versionadded: 0.3.0
+	.. versionadded: 0.4.0
 	"""
 
 	def __init__(
@@ -163,11 +170,11 @@ class GitHubManager(RepoHelper):
 		:param org: Whether the repository should be created for the organisation set as ``username``,
 			or for the authenticated user (default).
 
-		.. versionchanged:: 0.3.0
+		.. versionchanged:: 0.4.0
 
 			Removed the ``verbose`` option. Provide it to the class constructor instead.
 
-		.. versionchanged:: 0.3.0  Added the ``org`` argument.
+		.. versionchanged:: 0.4.0  Added the ``org`` argument.
 		"""
 
 		with self.echo_rate_limit():
@@ -202,11 +209,11 @@ class GitHubManager(RepoHelper):
 		:param org: Whether the repository should be created for the organisation set as ``username``,
 			or for the authenticated user (default).
 
-		.. versionchanged:: 0.3.0
+		.. versionchanged:: 0.4.0
 
 			Removed the ``verbose`` option. Provide it to the class constructor instead.
 
-		.. versionchanged:: 0.3.0  Added the ``org`` argument.
+		.. versionchanged:: 0.4.0  Added the ``org`` argument.
 		"""
 
 		with self.echo_rate_limit():
@@ -234,7 +241,7 @@ class GitHubManager(RepoHelper):
 
 		:param org:
 
-		.. versionadded:: 0.3.0
+		.. versionadded:: 0.4.0
 		"""  # noqa: D400
 
 		user = self.github.get_user()
@@ -246,16 +253,26 @@ class GitHubManager(RepoHelper):
 			self.assert_matching_usernames(user)
 			return user
 
-	def secrets(self, org: bool = False) -> int:
+	def secrets(
+			self,
+			org: bool = False,
+			overwrite: Optional[bool] = None,
+			PYPI_TOKEN: Optional[str] = None,
+			ANACONDA_TOKEN: Optional[str] = None
+			) -> int:
 		"""
 		Set or update the secrets of the GitHub repository for this project.
 
 		:param org: Whether the repository should be created for the organisation set as ``username``,
 			or for the authenticated user (default).
+		:param overwrite: Overwrite existing values.
+		:default overwrite: ask first.
 
-		.. versionadded:: 0.3.0
+		``PYPI_TOKEN`` and ``ANACONDA_TOKEN`` can either be passed as keyword arguments to this function or provided at the interactive prompt.
 
-		.. versionchanged:: 0.3.0  Added the ``org`` argument.
+		.. versionadded:: 0.4.0
+
+		.. versionchanged:: 0.4.0  Add ``overwrite``, ``PYPI_TOKEN``, ``ANACONDA_TOKEN`` options.
 		"""
 
 		with self.echo_rate_limit():
@@ -286,16 +303,22 @@ class GitHubManager(RepoHelper):
 				target_secrets.add("ANACONDA_TOKEN")
 
 			for secret_name in sorted(target_secrets):
-				if secret_name in existing_secrets:
+				if overwrite is not None:
+					update = True
+				elif secret_name in existing_secrets:
 					click.echo(f"A value for the secret {secret_name!r} already exists. ")
 					update = confirm("Do you want to update the secret?")
 				else:
 					update = True
 
 				if update:
-					operation = "create" if secret_name in existing_secrets else "update"
+					operation = "update" if secret_name in existing_secrets else "create"
+
+					encrypted_value = encrypt_secret(
+							public_key["key"],
+							secret_value=locals().get(secret_name, None) or getpass(f"{secret_name}: "),
+							)
 
-					encrypted_value = encrypt_secret(public_key["key"], getpass(f"{secret_name}: "))
 					key_id = public_key["key_id"]
 					secret_json = {"encrypted_value": encrypted_value, "key_id": key_id}
 					response = (secrets_url / secret_name).put(json=secret_json)
@@ -310,6 +333,43 @@ class GitHubManager(RepoHelper):
 
 		return ret
 
+	def protect_branch(self, branch: str, org: bool = False) -> int:
+		"""
+		Update branch protection for the given branch.
+
+		This requires that the Linux and Windows tests pass, together with the mypy check.
+
+		:param branch: The branch to update protection for.
+		:param org: Whether the repository should be created for the organisation set as ``username``,
+			or for the authenticated user (default).
+
+		.. versionadded:: 0.4.0
+		"""
+
+		with self.echo_rate_limit():
+			user = self.get_org_or_user(org)
+
+			try:
+				repo: Repository = user.get_repo(self.templates.globals["repo_name"])
+			except GithubException as e:
+				self.handle_exception(e)
+
+			gh_branch = repo.get_branch(branch)
+			required_checks = list(compile_required_checks(self))
+
+			# gh_branch.edit_required_status_checks(strict=False, contexts=)
+			gh_branch.edit_protection(
+					strict=False,
+					contexts=required_checks,
+					dismiss_stale_reviews=False,
+					required_approving_review_count=1,
+					)
+
+			gh_branch.edit_protection(strict=False, contexts=required_checks)
+
+		click.echo("Up to date!")
+		return 0
+
 	def assert_matching_usernames(self, user: AuthenticatedUser):
 		"""
 		Assert that the username configured in ``repo_helper.yml`` matches that of the authenticated user.
@@ -363,7 +423,12 @@ class GitHubManager(RepoHelper):
 		"""
 
 		data: _ExcData = exc.data  # type: ignore
-		errors = DelimitedList(i["message"] for i in data["errors"])
+
+		if "errors" in data:
+			errors = DelimitedList(i["message"] for i in data["errors"])
+		else:
+			errors = DelimitedList([data["message"]])
+
 		raise abort(f"{exc.data['message']}\n{errors:\t\n}")
 
 	def get_repo_kwargs(self) -> _EditKwargs:
@@ -371,7 +436,12 @@ class GitHubManager(RepoHelper):
 		Returns the keyword arguments used when creating and updating repositories.
 		"""
 
-		edit_kwargs: _EditKwargs = {"description": self.templates.globals["short_desc"]}
+		# TODO: add config option for allow_merge_commit
+
+		edit_kwargs: _EditKwargs = {
+				"description": self.templates.globals["short_desc"],
+				"allow_merge_commit": False,
+				}
 
 		if self.templates.globals["enable_docs"]:
 			edit_kwargs["homepage"] = self.templates.globals["docs_url"]
@@ -386,7 +456,7 @@ def encrypt_secret(public_key: str, secret_value: str) -> str:
 	:param public_key:
 	:param secret_value:
 
-	.. versionadded: 0.3.0
+	.. versionadded: 0.4.0
 	"""
 
 	public_key = public.PublicKey(public_key.encode("utf-8"), encoding.Base64Encoder())
@@ -395,4 +465,89 @@ def encrypt_secret(public_key: str, secret_value: str) -> str:
 	return b64encode(encrypted).decode("utf-8")
 
 
+def compile_required_checks(repo: RepoHelper) -> Iterator[str]:
+	"""
+	Returns an iterator over the names of required checks for the given repository.
+
+	:param repo:
+
+	.. versionadded:: 0.4.0
+	"""
+
+	actions_manager = ActionsManager(repo.target_repo, repo.templates)
+
+	for platform in repo.templates.globals["platforms"]:
+		if platform not in platform_ci_names:
+			continue
+
+		if platform == "Windows":
+			ci_platform = platform_ci_names[platform]
+			py_versions = actions_manager.get_windows_ci_versions()
+		elif platform == "Linux":
+			ci_platform = platform_ci_names[platform]
+			py_versions = actions_manager.get_linux_ci_versions()
+		# elif platform == "macOS":
+		# 	ci_platform = platform_ci_names[platform]
+		# 	py_versions = actions_manager.get_macos_ci_versions()
+		else:
+			continue
+
+		for version in set_gh_actions_versions(py_versions):
+			if "alpha" in version or "beta" in version or "-dev" in version:
+				continue
+
+			yield f"{ci_platform} / Python {version}"
+
+	yield from ["mypy / ubuntu-latest", "Flake8"]
+
+	if repo.templates.globals["enable_tests"]:
+		yield "docs"
+
+
 GithubManager = GitHubManager
+
+
+class IsolatedGitHubManager(GitHubManager):
+	"""
+	Subclass of :class:`~.GitHubManager` which can be used isolated from the repository
+	and its ``repo_helper.yml`` config file.
+
+	:param token: The token to authenticate with the GitHub API.
+
+	:param username: The username of the GitHub account hosting the repository.
+	:param repo_name: The name of GitHub repository.
+	:param managed_message: Message placed at the top of files to indicate that they are managed by ``repo_helper``.
+	:param verbose: Whether to show information on the GitHub API rate limit.
+	:param colour: Whether to use coloured output.
+
+	.. versionadded:: 0.4.0
+	"""  # noqa: D400
+
+	def __init__(
+			self,
+			token: str,
+			username: str,
+			repo_name: str,
+			*,
+			managed_message="This file is managed by 'repo_helper'. Don't edit it directly.",
+			verbose: bool = False,
+			colour: Optional[bool] = True,
+			):
+
+		self._tmpdir = tempfile.TemporaryDirectory()
+
+		self.github = Github(token)
+		self.verbose = verbose
+		self.colour = resolve_color_default(colour)
+
+		target_repo = PathPlus(self._tmpdir.name)
+		config_file_name = "repo_helper.yml"
+
+		github_repo = self.github.get_repo(f"{username}/{repo_name}")
+		contents_from_github: ContentFile = github_repo.get_contents(config_file_name)  # type: ignore
+		(target_repo / config_file_name).write_bytes(contents_from_github.decoded_content)
+
+		RepoHelper.__init__(self, target_repo, managed_message)
+
+	def __del__(self):
+		self._tmpdir.cleanup()
